@@ -1,4 +1,4 @@
-
+from pyopencl.version import VERSION, VERSION_STATUS, VERSION_TEXT
 
 try:
     import pyopencl._cl as _cl
@@ -37,6 +37,14 @@ def compiler_output(text):
 
 
 
+# {{{ Kernel
+
+class Kernel(_cl._Kernel):
+    def __init__(self, prg, name):
+        _cl._Kernel.__init__(self, prg._get_prg(), name)
+
+# }}}
+
 # {{{ Program (including caching support)
 
 class Program(object):
@@ -89,7 +97,7 @@ class Program(object):
 
     def __getattr__(self, attr):
         try:
-            knl = Kernel(self._get_prg(), attr)
+            knl = Kernel(self, attr)
             # Nvidia does not raise errors even for invalid names,
             # but this will give an error if the kernel is invalid.
             knl.num_args
@@ -146,7 +154,7 @@ class Image(_cl._ImageBase):
         if shape is None and hostbuf is not None:
             shape = hostbuf.shape
 
-        if hostbuf is None and not \
+        if hostbuf is not None and not \
                 (flags & (mem_flags.USE_HOST_PTR | mem_flags.COPY_HOST_PTR)):
             from warnings import warn
             warn("'hostbuf' was passed, but no memory flags to make use of it.")
@@ -261,7 +269,7 @@ def _add_functionality():
                 (_cl._ImageBase.get_image_info, _cl.image_info),
             Program:
                 (Program.get_info, _cl.program_info),
-            _cl.Kernel:
+            Kernel:
                 (Kernel.get_info, _cl.kernel_info),
             _cl.Sampler:
                 (Sampler.get_info, _cl.sampler_info),
@@ -481,7 +489,7 @@ def _add_functionality():
                 for i, arg in enumerate(args):
                     self.set_arg(i, arg)
             else:
-                from struct import pack
+                from pyopencl._pvt_struct import pack
 
                 for i, (arg, arg_type_char) in enumerate(
                         zip(args, arg_type_chars)):
@@ -687,6 +695,8 @@ def create_some_context(interactive=True, answers=None):
                 "avoid being asked again." % ":".join(user_inputs))
     return Context(devices)
 
+_csc = create_some_context
+
 
 
 
