@@ -30,17 +30,90 @@ OTHER DEALINGS IN THE SOFTWARE.
 
 
 
-import numpy
+import numpy as np
 import pyopencl.elementwise as elementwise
 import pyopencl as cl
-#from pytools import memoize_method
+from pytools import memoize_method
+from pyopencl.compyte.array import (
+        as_strided as _as_strided,
+        f_contiguous_strides as _f_contiguous_strides,
+        c_contiguous_strides as _c_contiguous_strides,
+        ArrayFlags as _ArrayFlags,
+        get_common_dtype as _get_common_dtype_base)
+from pyopencl.characterize import has_double_support
 
 
 
-def splay(queue, n):
+
+def _get_common_dtype(obj1, obj2, queue):
+    return _get_common_dtype_base(obj1, obj2,
+            has_double_support(queue.device))
+
+
+
+# {{{ vector types
+
+class vec:
+    pass
+
+def _create_vector_types():
+    field_names = ["x", "y", "z", "w"]
+
+    from pyopencl.tools import register_dtype
+
+    counts = [2, 3, 4, 8, 16]
+    for base_name, base_type in [
+        ('char', np.int8),
+        ('uchar', np.uint8),
+        ('short', np.int16),
+        ('ushort', np.uint16),
+        ('int', np.int32),
+        ('uint', np.uint32),
+        ('long', np.int64),
+        ('ulong', np.uint64),
+        ('float', np.float32),
+        ('double', np.float64),
+        ]:
+        for count in counts:
+            name = "%s%d" % (base_name, count)
+
+            titles = field_names[:count]
+            if len(titles) < count:
+                titles.extend((count-len(titles))*[None])
+
+            dtype = np.dtype(dict(
+                names=["s%d" % i for i in range(count)],
+                formats=[base_type]*count,
+                titles=titles))
+
+            register_dtype(dtype, name)
+
+            setattr(vec, name, dtype)
+
+            my_field_names = ",".join(field_names[:count])
+            my_field_names_defaulted = ",".join(
+                    "%s=0" % fn for fn in field_names[:count])
+            setattr(vec, "make_"+name,
+                    staticmethod(eval(
+                        "lambda %s: array((%s), dtype=my_dtype)"
+                        % (my_field_names_defaulted, my_field_names),
+                        dict(array=np.array, my_dtype=dtype))))
+
+_create_vector_types()
+
+# }}}
+
+# {{{ helper functionality
+
+def splay(queue, n, kernel_specific_max_wg_size=None):
     dev = queue.device
-    max_work_items = min(128, dev.max_work_group_size)
-    min_work_items = min(32, max_work_items)
+    max_work_items = _builtin_min(128, dev.max_work_group_size)
+
+    if kernel_specific_max_wg_size is not None:
+        from __builtin__ import min
+        max_work_items = min(max_work_items, kernel_specific_max_wg_size)
+
+    min_work_items = _builtin_min(32, max_work_items)
     max_groups = dev.max_compute_units * 4 * 8
     # 4 to overfill the device
     # 8 is an Nvidia constant--that's how many
@@ -67,12 +140,6 @@ def splay(queue, n):
 
 
 
-def _get_common_dtype(obj1, obj2):
-    return (obj1.dtype.type(0) + obj2.dtype.type(0)).dtype
-
-
-
-
 def elwise_kernel_runner(kernel_getter):
     """Take a kernel getter of the same signature as the kernel
     and return a function that invokes that kernel.
@@ -85,14 +152,21 @@ def elwise_kernel_runner(kernel_getter):
         repr_ary = args[0]
         queue = kwargs.pop("queue", None) or repr_ary.queue
 
-        gs, ls = repr_ary.get_sizes(queue)
         knl = kernel_getter(*args)
 
+        gs, ls = repr_ary.get_sizes(queue,
+                knl.get_work_group_info(
+                    cl.kernel_work_group_info.WORK_GROUP_SIZE,
+                    queue.device))
+
         assert isinstance(repr_ary, Array)
 
         actual_args = []
         for arg in args:
             if isinstance(arg, Array):
+                if not arg.flags.forc:
+                    raise RuntimeError("only contiguous arrays may "
+                            "be used as arguments to this operation")
                 actual_args.append(arg.data)
             else:
                 actual_args.append(arg)
@@ -100,24 +174,36 @@ def elwise_kernel_runner(kernel_getter):
 
         return knl(queue, gs, ls, *actual_args)
 
-    from functools import update_wrapper
-    return update_wrapper(kernel_runner, kernel_getter)
+    try:
+       from functools import update_wrapper
+    except ImportError:
+        return kernel_runner
+    else:
+       return update_wrapper(kernel_runner, kernel_getter)
 
 
 
 
 
-class DefaultAllocator:
-    def __init__(self, context, flags=cl.mem_flags.READ_WRITE):
-        self.context = context
-        self.flags = flags
 
-    def __call__(self, size):
-        return cl.Buffer(self.context, self.flags, size)
+DefaultAllocator = cl.CLAllocator
 
 
 
 
+def _should_be_cqa(what):
+    from warnings import warn
+    warn("'%s' should be specified as the frst"
+            "('cqa') argument, "
+            "not in the '%s' keyword argument. "
+            "This will be continue to be accepted througout "
+            "versions 2011.x of PyOpenCL." % (what, what),
+            DeprecationWarning, 3)
+
+# }}}
+
+# {{{ array class
+
 class Array(object):
     """A :mod:`pyopencl` Array is used to do array-based calculation on
     a compute device.
@@ -126,30 +212,82 @@ class Array(object):
     work on an element-by-element basis, just like :class:`numpy.ndarray`.
     """
 
-    def __init__(self, context, shape, dtype, order="C", allocator=None,
-            base=None, data=None, queue=None):
-        if allocator is None:
-            allocator = DefaultAllocator(context)
+    def __init__(self, cqa, shape, dtype, order="C", allocator=None,
+            base=None, data=None, queue=None, strides=None):
+        # {{{ backward compatibility for pre-cqa days
+
+        if isinstance(cqa, cl.CommandQueue):
+            if queue is not None:
+                raise TypeError("can't specify queue in 'cqa' and "
+                        "'queue' arguments")
+            queue = cqa
+
+            if allocator is None:
+                context = queue.context
+                allocator = DefaultAllocator(context)
+
+        elif isinstance(cqa, cl.Context):
+            if queue is not None:
+                _should_be_cqa("queue")
+
+            if allocator is not None:
+                _should_be_cqa("allocator")
+            else:
+                allocator = DefaultAllocator(cqa)
+
+        else:
+            # cqa is assumed to be an allocator
+            if allocator is not None:
+                raise TypeError("can't specify allocator in 'cqa' and "
+                        "'allocator' arguments")
+
+            allocator = cqa
+
+        # }}}
+
+        # invariant here: allocator, queue set
+
+        # {{{ determine shape and strides
+        dtype = np.dtype(dtype)
 
         try:
             s = 1
             for dim in shape:
                 s *= dim
         except TypeError:
-            if not isinstance(shape, int):
+            import sys
+            if sys.version_info >= (3,):
+                admissible_types = (int, np.integer)
+            else:
+                admissible_types = (int, long, np.integer)
+
+            if not isinstance(shape, admissible_types):
                 raise TypeError("shape must either be iterable or "
                         "castable to an integer")
             s = shape
             shape = (shape,)
 
-        self.context = context
-        self.queue = queue
+        if strides is None:
+            if order == "F":
+                strides = _f_contiguous_strides(
+                        dtype.itemsize, shape)
+            elif order == "C":
+                strides = _c_contiguous_strides(
+                        dtype.itemsize, shape)
+            else:
+                raise ValueError("invalid order: %s" % order)
+        else:
+            # FIXME: We should possibly perform some plausibility
+            # checking on 'strides' here.
 
+            strides = tuple(strides)
+
+        # }}}
+
+        self.queue = queue
         self.shape = shape
-        self.dtype = numpy.dtype(dtype)
-        if order not in ["C", "F"]:
-            raise ValueError("order must be either 'C' or 'F'")
-        self.order = order
+        self.dtype = dtype
+        self.strides = strides
 
         self.mem_size = self.size = s
         self.nbytes = self.dtype.itemsize * self.size
@@ -160,41 +298,82 @@ class Array(object):
                 self.data = self.allocator(self.size * self.dtype.itemsize)
             else:
                 self.data = None
-
-            if base is not None:
-                raise ValueError("If data is specified, base must be None.")
         else:
             self.data = data
 
         self.base = base
 
+    @property
+    def context(self):
+        return self.data.context
+
+    @property
+    @memoize_method
+    def flags(self):
+        return _ArrayFlags(self)
+
+    def _new_with_changes(self, data, shape=None, dtype=None, strides=None, queue=None,
+            base=None):
+        if shape is None:
+            shape = self.shape
+        if dtype is None:
+            dtype = self.dtype
+        if strides is None:
+            strides = self.strides
+        if queue is None:
+            queue = self.queue
+        if base is None and data is self.data:
+            base = self
+
+        if queue is not None:
+            return Array(queue, shape, dtype,
+                    allocator=self.allocator, strides=strides, base=base)
+        elif self.allocator is not None:
+            return Array(self.allocator, shape, dtype, queue=queue,
+                    strides=strides, base=base)
+        else:
+            return Array(self.context, shape, dtype, strides=strides, base=base)
+
     #@memoize_method FIXME: reenable
-    def get_sizes(self, queue):
-        return splay(queue, self.mem_size)
+    def get_sizes(self, queue, kernel_specific_max_wg_size=None):
+        return splay(queue, self.mem_size,
+                kernel_specific_max_wg_size=kernel_specific_max_wg_size)
 
     def set(self, ary, queue=None, async=False):
         assert ary.size == self.size
         assert ary.dtype == self.dtype
-        if self.size:
-            evt = cl.enqueue_write_buffer(queue or self.queue, self.data, ary)
 
-            if not async:
-                evt.wait()
+        if not ary.flags.forc:
+            raise RuntimeError("cannot set from non-contiguous array")
+
+            ary = ary.copy()
+
+        if ary.strides != self.strides:
+            from warnings import warn
+            warn("Setting array from one with different strides/storage order. "
+                    "This will cease to work in 2013.x.",
+                    stacklevel=2)
+
+        if self.size:
+            cl.enqueue_copy(queue or self.queue, self.data, ary,
+                    is_blocking=not async)
 
     def get(self, queue=None, ary=None, async=False):
         if ary is None:
-            ary = numpy.empty(self.shape, self.dtype, order=self.order)
+            ary = np.empty(self.shape, self.dtype)
+
+            ary = _as_strided(ary, strides=self.strides)
         else:
             if ary.size != self.size:
                 raise TypeError("'ary' has non-matching type")
             if ary.dtype != self.dtype:
                 raise TypeError("'ary' has non-matching size")
 
-        if self.size:
-            evt = cl.enqueue_read_buffer(queue or self.queue, self.data, ary)
+        assert self.flags.forc, "Array in get() must be contiguous"
 
-            if not async:
-                evt.wait()
+        if self.size:
+            cl.enqueue_copy(queue or self.queue, ary, self.data,
+                    is_blocking=not async)
 
         return ary
 
@@ -220,9 +399,12 @@ class Array(object):
 
     @staticmethod
     @elwise_kernel_runner
-    def _axpbz(out, afac, a, other, queue=None):
-        """Compute ``out = afac * a + other``, where `other` is a scalar."""
-        return elementwise.get_axpbz_kernel(out.context, out.dtype)
+    def _axpbz(out, a, x, b, queue=None):
+        """Compute ``z = a * x + b``, where `b` is a scalar."""
+        a = np.array(a)
+        b = np.array(b)
+        return elementwise.get_axpbz_kernel(out.context,
+                a.dtype, x.dtype, b.dtype, out.dtype)
 
     @staticmethod
     @elwise_kernel_runner
@@ -233,12 +415,13 @@ class Array(object):
     @staticmethod
     @elwise_kernel_runner
     def _rdiv_scalar(out, ary, other, queue=None):
+        other = np.array(other)
         return elementwise.get_rdivide_elwise_kernel(
-                out.context, ary.dtype)
+                out.context, ary.dtype, other.dtype, out.dtype)
 
     @staticmethod
     @elwise_kernel_runner
-    def _div(self, out, other, queue=None):
+    def _div(out, self, other, queue=None):
         """Divides an array by another array."""
 
         assert self.shape == other.shape
@@ -254,7 +437,10 @@ class Array(object):
     @staticmethod
     @elwise_kernel_runner
     def _abs(result, arg):
-        if arg.dtype.kind == "f":
+        if arg.dtype.kind == "c":
+            from pyopencl.elementwise import complex_dtype_to_name
+            fname = "%s_abs" % complex_dtype_to_name(arg.dtype)
+        elif arg.dtype.kind == "f":
             fname = "fabs"
         elif arg.dtype.kind in ["u", "i"]:
             fname = "abs"
@@ -262,18 +448,54 @@ class Array(object):
             raise TypeError("unsupported dtype in _abs()")
 
         return elementwise.get_unary_func_kernel(
-                arg.context, fname, arg.dtype)
+                arg.context, fname, arg.dtype, out_dtype=result.dtype)
+
+    @staticmethod
+    @elwise_kernel_runner
+    def _real(result, arg):
+        from pyopencl.elementwise import complex_dtype_to_name
+        fname = "%s_real" % complex_dtype_to_name(arg.dtype)
+        return elementwise.get_unary_func_kernel(
+                arg.context, fname, arg.dtype, out_dtype=result.dtype)
+
+    @staticmethod
+    @elwise_kernel_runner
+    def _imag(result, arg):
+        from pyopencl.elementwise import complex_dtype_to_name
+        fname = "%s_imag" % complex_dtype_to_name(arg.dtype)
+        return elementwise.get_unary_func_kernel(
+                arg.context, fname, arg.dtype, out_dtype=result.dtype)
+
+    @staticmethod
+    @elwise_kernel_runner
+    def _conj(result, arg):
+        from pyopencl.elementwise import complex_dtype_to_name
+        fname = "%s_conj" % complex_dtype_to_name(arg.dtype)
+        return elementwise.get_unary_func_kernel(
+                arg.context, fname, arg.dtype, out_dtype=result.dtype)
 
     @staticmethod
     @elwise_kernel_runner
     def _pow_scalar(result, ary, exponent):
-        return elementwise.get_pow_kernel(result.context, result.dtype)
+        exponent = np.array(exponent)
+        return elementwise.get_pow_kernel(result.context,
+                ary.dtype, exponent.dtype, result.dtype,
+                is_base_array=True, is_exp_array=False)
+
+    @staticmethod
+    @elwise_kernel_runner
+    def _rpow_scalar(result, base, exponent):
+        base = np.array(base)
+        return elementwise.get_pow_kernel(result.context,
+                base.dtype, exponent.dtype, result.dtype,
+                is_base_array=False, is_exp_array=True)
 
     @staticmethod
     @elwise_kernel_runner
     def _pow_array(result, base, exponent):
-        return elementwise.get_pow_array_kernel(
-                result.context, base.dtype, exponent.dtype, result.dtype)
+        return elementwise.get_pow_kernel(
+                result.context, base.dtype, exponent.dtype, result.dtype,
+                is_base_array=True, is_exp_array=True)
 
     @staticmethod
     @elwise_kernel_runner
@@ -287,17 +509,30 @@ class Array(object):
                 dest.context, dest.dtype, src.dtype)
 
     def _new_like_me(self, dtype=None, queue=None):
+        strides = None
         if dtype is None:
             dtype = self.dtype
-        return self.__class__(self.context,
-                self.shape, dtype, allocator=self.allocator,
-                queue=queue or self.queue)
+        else:
+            if dtype == self.dtype:
+                strides = self.strides
+
+        queue = queue or self.queue
+        if queue is not None:
+            return self.__class__(queue, self.shape, dtype,
+                    allocator=self.allocator, strides=strides)
+        elif self.allocator is not None:
+            return self.__class__(self.allocator, self.shape, dtype,
+                    strides=strides)
+        else:
+            return self.__class__(self.context, self.shape, dtype,
+                    strides=strides)
 
     # operators ---------------------------------------------------------------
     def mul_add(self, selffac, other, otherfac, queue=None):
         """Return `selffac * self + otherfac*other`.
         """
-        result = self._new_like_me(_get_common_dtype(self, other))
+        result = self._new_like_me(
+                _get_common_dtype(self, other, queue or self.queue))
         self._axpbyz(result, selffac, self, otherfac, other)
         return result
 
@@ -306,16 +541,19 @@ class Array(object):
 
         if isinstance(other, Array):
             # add another vector
-            result = self._new_like_me(_get_common_dtype(self, other))
-            self._axpbyz(result, 1, self, 1, other)
+            result = self._new_like_me(_get_common_dtype(self, other, self.queue))
+            self._axpbyz(result,
+                    self.dtype.type(1), self,
+                    other.dtype.type(1), other)
             return result
         else:
             # add a scalar
             if other == 0:
                 return self
             else:
-                result = self._new_like_me()
-                self._axpbz(result, 1, self, other)
+                common_dtype = _get_common_dtype(self, other, self.queue)
+                result = self._new_like_me(common_dtype)
+                self._axpbz(result, self.dtype.type(1), self, common_dtype.type(other))
                 return result
 
     __radd__ = __add__
@@ -324,16 +562,18 @@ class Array(object):
         """Substract an array from an array or a scalar from an array."""
 
         if isinstance(other, Array):
-            result = self._new_like_me(_get_common_dtype(self, other))
-            self._axpbyz(result, 1, self, -1, other)
+            result = self._new_like_me(_get_common_dtype(self, other, self.queue))
+            self._axpbyz(result,
+                    self.dtype.type(1), self,
+                    other.dtype.type(-1), other)
             return result
         else:
             # subtract a scalar
             if other == 0:
                 return self
             else:
-                result = self._new_like_me()
-                self._axpbz(result, 1, self, -other)
+                result = self._new_like_me(_get_common_dtype(self, other, self.queue))
+                self._axpbz(result, self.dtype.type(1), self, -other)
                 return result
 
     def __rsub__(self,other):
@@ -341,25 +581,28 @@ class Array(object):
 
            x = n - self
         """
+        common_dtype = _get_common_dtype(self, other, self.queue)
         # other must be a scalar
-        result = self._new_like_me()
-        self._axpbz(result, -1, self, other)
+        result = self._new_like_me(common_dtype)
+        self._axpbz(result, self.dtype.type(-1), self, common_dtype.type(other))
         return result
 
     def __iadd__(self, other):
         if isinstance(other, Array):
-            self._axpbyz(self, 1, self, 1, other)
+            self._axpbyz(self,
+                    self.dtype.type(1), self,
+                    other.dtype.type(1), other)
             return self
         else:
-            self._axpbz(self, 1, self, other)
+            self._axpbz(self, self.dtype.type(1), self, other)
             return self
 
     def __isub__(self, other):
         if isinstance(other, Array):
-            self._axpbyz(self, 1, self, -1, other)
+            self._axpbyz(self, self.dtype.type(1), self, other.dtype.type(-1), other)
             return self
         else:
-            self._axpbz(self, 1, self, -other)
+            self._axpbz(self, self.dtype.type(1), self, -other)
             return self
 
     def __neg__(self):
@@ -369,21 +612,28 @@ class Array(object):
 
     def __mul__(self, other):
         if isinstance(other, Array):
-            result = self._new_like_me(_get_common_dtype(self, other))
+            result = self._new_like_me(_get_common_dtype(self, other, self.queue))
             self._elwise_multiply(result, self, other)
             return result
         else:
-            result = self._new_like_me()
-            self._axpbz(result, other, self, 0)
+            common_dtype = _get_common_dtype(self, other, self.queue)
+            result = self._new_like_me(common_dtype)
+            self._axpbz(result, common_dtype.type(other), self, self.dtype.type(0))
             return result
 
     def __rmul__(self, scalar):
-        result = self._new_like_me()
-        self._axpbz(result, scalar, self, 0)
+        common_dtype = _get_common_dtype(self, scalar, self.queue)
+        result = self._new_like_me(common_dtype)
+        self._axpbz(result, common_dtype.type(scalar), self, self.dtype.type(0))
         return result
 
-    def __imul__(self, scalar):
-        self._axpbz(self, scalar, self, 0)
+    def __imul__(self, other):
+        if isinstance(other, Array):
+            self._elwise_multiply(self, self, other)
+        else:
+            # scalar
+            self._axpbz(self, other, self, self.dtype.type(0))
+
         return self
 
     def __div__(self, other):
@@ -392,15 +642,17 @@ class Array(object):
            x = self / n
         """
         if isinstance(other, Array):
-            result = self._new_like_me(_get_common_dtype(self, other))
+            result = self._new_like_me(_get_common_dtype(self, other, self.queue))
             self._div(result, self, other)
         else:
             if other == 1:
                 return self
             else:
                 # create a new array for the result
-                result = self._new_like_me()
-                self._axpbz(result, 1/other, self, 0)
+                common_dtype = _get_common_dtype(self, other, self.queue)
+                result = self._new_like_me(common_dtype)
+                self._axpbz(result,
+                        common_dtype.type(1/other), self, self.dtype.type(0))
 
         return result
 
@@ -413,19 +665,20 @@ class Array(object):
         """
 
         if isinstance(other, Array):
-            result = self._new_like_me(_get_common_dtype(self, other))
+            result = self._new_like_me(_get_common_dtype(self, other, self.queue))
             other._div(result, self)
         else:
             if other == 1:
                 return self
             else:
                 # create a new array for the result
-                result = self._new_like_me()
-                self._rdiv_scalar(result, self, other)
+                common_dtype = _get_common_dtype(self, other, self.queue)
+                result = self._new_like_me(common_dtype)
+                self._rdiv_scalar(result, self, common_dtype.type(other))
 
         return result
 
-    __rtruediv__ = __div__
+    __rtruediv__ = __rdiv__
 
     def fill(self, value, queue=None):
         """fills the array with the specified value"""
@@ -444,7 +697,7 @@ class Array(object):
         of `self`.
         """
 
-        result = self._new_like_me()
+        result = self._new_like_me(self.dtype.type(0).real.dtype)
         self._abs(result, self)
         return result
 
@@ -456,14 +709,21 @@ class Array(object):
         if isinstance(other, Array):
             assert self.shape == other.shape
 
-            result = self._new_like_me(_get_common_dtype(self, other))
+            result = self._new_like_me(_get_common_dtype(self, other, self.queue))
             self._pow_array(result, self, other)
         else:
-            result = self._new_like_me()
+            result = self._new_like_me(_get_common_dtype(self, other, self.queue))
             self._pow_scalar(result, self, other)
 
         return result
 
+    def __rpow__(self, other):
+        # other must be a scalar
+        common_dtype = _get_common_dtype(self, other, self.queue)
+        result = self._new_like_me(common_dtype)
+        self._rpow_scalar(result, common_dtype.type(other), self)
+        return result
+
     def reverse(self, queue=None):
         """Return this array in reversed order. The array is treated
         as one-dimensional.
@@ -500,60 +760,136 @@ class Array(object):
     def __gt__(self, other):
         raise NotImplementedError
 
+    # {{{ complex-valued business
 
+    @property
+    def real(self):
+        if self.dtype.kind == "c":
+            result = self._new_like_me(self.dtype.type(0).real.dtype)
+            self._real(result, self)
+            return result
+        else:
+            return self
 
-def to_device(context, queue, ary, allocator=None, async=False):
+    @property
+    def imag(self):
+        if self.dtype.kind == "c":
+            result = self._new_like_me(self.dtype.type(0).real.dtype)
+            self._imag(result, self)
+            return result
+        else:
+            return zeros_like(self)
+
+    def conj(self):
+        if self.dtype.kind == "c":
+            result = self._new_like_me()
+            self._conj(result, self)
+            return result
+        else:
+            return self
+
+    # }}}
+
+    # {{{ views
+
+    def reshape(self, *shape):
+        # TODO: add more error-checking, perhaps
+        if isinstance(shape[0], tuple) or isinstance(shape[0], list):
+            shape = tuple(shape[0])
+        size = reduce(lambda x, y: x * y, shape, 1)
+        if size != self.size:
+            raise ValueError("total size of new array must be unchanged")
+
+        return self._new_with_changes(data=self.data, shape=shape)
+
+    def ravel(self):
+        return self.reshape(self.size)
+
+    def view(self, dtype=None):
+        if dtype is None:
+            dtype = self.dtype
+
+        old_itemsize = self.dtype.itemsize
+        itemsize = np.dtype(dtype).itemsize
+
+        if self.shape[-1] * old_itemsize % itemsize != 0:
+            raise ValueError("new type not compatible with array")
+
+        shape = self.shape[:-1] + (self.shape[-1] * old_itemsize // itemsize,)
+
+        return self._new_with_changes(data=self.data, shape=shape, dtype=dtype)
+
+    # }}
+
+# }}}
+
+# }}}
+
+# {{{ creation helpers
+
+def to_device(*args, **kwargs):
     """Converts a numpy array to a :class:`Array`."""
-    if ary.flags.f_contiguous:
-        order = "F"
-    elif ary.flags.c_contiguous:
-        order = "C"
-    else:
-        raise ValueError("to_device only works on C- or Fortran-"
-                "contiguous arrays")
 
-    result = Array(context, ary.shape, ary.dtype, order, allocator,
-            queue=queue)
-    result.set(ary, async=async)
-    return result
+    def _to_device(queue, ary, allocator=None, async=False):
+        if ary.dtype == object:
+            raise RuntimeError("to_device does not work on object arrays.")
+
+        result = Array(queue, ary.shape, ary.dtype,
+                        allocator=allocator, strides=ary.strides)
+        result.set(ary, async=async)
+        return result
+
+    if isinstance(args[0], cl.Context):
+        from warnings import warn
+        warn("Passing a context as first argument is deprecated. "
+            "This will be continue to be accepted througout "
+            "versions 2011.x of PyOpenCL.",
+            DeprecationWarning, 2)
+        args = args[1:]
+
+    return _to_device(*args, **kwargs)
+
+
 
 
 empty = Array
 
-def zeros(context, queue, shape, dtype, order="C", allocator=None):
+def zeros(*args, **kwargs):
     """Returns an array of the given shape and dtype filled with 0's."""
 
-    result = Array(context, shape, dtype, 
-            order=order, allocator=allocator, queue=queue)
-    result.fill(0)
-    return result
+    def _zeros(queue, shape, dtype, order="C", allocator=None):
+        result = Array(queue, shape, dtype,
+                order=order, allocator=allocator)
+        zero = np.zeros((), dtype)
+        result.fill(zero)
+        return result
+
+    if isinstance(args[0], cl.Context):
+        from warnings import warn
+        warn("Passing a context as first argument is deprecated. "
+            "This will be continue to be accepted througout "
+            "versions 2011.x of PyOpenCL.",
+            DeprecationWarning, 2)
+        args = args[1:]
+
+    return _zeros(*args, **kwargs)
 
 def empty_like(ary):
-    result = Array(ary.context,
-            ary.shape, ary.dtype, allocator=ary.allocator, queue=ary.queue)
-    return result
+    return ary._new_with_changes(data=None)
 
 def zeros_like(ary):
-    result = Array(ary.context,
-            ary.shape, ary.dtype, allocator=ary.allocator, queue=ary.queue)
-    result.fill(0)
+    result = empty_like(ary)
+    zero = np.zeros((), ary.dtype)
+    result.fill(zero)
     return result
 
 
 @elwise_kernel_runner
-def _arange(result, start, step):
+def _arange_knl(result, start, step):
     return elementwise.get_arange_kernel(
             result.context, result.dtype)
 
-def arange(context, queue, *args, **kwargs):
-    """Create an array filled with numbers spaced `step` apart,
-    starting from `start` and ending at `stop`.
-
-    For floating point arguments, the length of the result is
-    `ceil((stop - start)/step)`.  This rule may result in the last
-    element of the result being greater than stop.
-    """
-
+def _arange(queue, *args, **kwargs):
     # argument processing -----------------------------------------------------
 
     # Yuck. Thanks, numpy developers. ;)
@@ -568,8 +904,9 @@ def arange(context, queue, *args, **kwargs):
     inf.stop = None
     inf.step = None
     inf.dtype = None
+    inf.allocator = None
 
-    if isinstance(args[-1], numpy.dtype):
+    if isinstance(args[-1], np.dtype):
         dtype = args[-1]
         args = args[:-1]
         explicit_dtype = True
@@ -589,7 +926,7 @@ def arange(context, queue, *args, **kwargs):
     else:
         raise ValueError, "too many arguments"
 
-    admissible_names = ["start", "stop", "step", "dtype"]
+    admissible_names = ["start", "stop", "step", "dtype", "allocator"]
     for k, v in kwargs.iteritems():
         if k in admissible_names:
             if getattr(inf, k) is None:
@@ -606,10 +943,10 @@ def arange(context, queue, *args, **kwargs):
     if inf.step is None:
         inf.step = 1
     if inf.dtype is None:
-        inf.dtype = numpy.array([inf.start, inf.stop, inf.step]).dtype
+        inf.dtype = np.array([inf.start, inf.stop, inf.step]).dtype
 
     # actual functionality ----------------------------------------------------
-    dtype = numpy.dtype(inf.dtype)
+    dtype = np.dtype(inf.dtype)
     start = dtype.type(inf.start)
     step = dtype.type(inf.step)
     stop = dtype.type(inf.stop)
@@ -620,13 +957,36 @@ def arange(context, queue, *args, **kwargs):
     from math import ceil
     size = int(ceil((stop-start)/step))
 
-    result = Array(context, (size,), dtype, queue=queue)
-    _arange(result, start, step, queue=queue)
+    result = Array(queue, (size,), dtype, allocator=inf.allocator)
+    _arange_knl(result, start, step, queue=queue)
     return result
 
 
 
 
+def arange(*args, **kwargs):
+    """Create an array filled with numbers spaced `step` apart,
+    starting from `start` and ending at `stop`.
+
+    For floating point arguments, the length of the result is
+    `ceil((stop - start)/step)`.  This rule may result in the last
+    element of the result being greater than stop.
+    """
+
+    if isinstance(args[0], cl.Context):
+        from warnings import warn
+        warn("Passing a context as first argument is deprecated. "
+            "This will be continue to be accepted througout "
+            "versions 2011.x of PyOpenCL.",
+            DeprecationWarning, 2)
+        args = args[1:]
+
+    return _arange(*args, **kwargs)
+
+# }}}
+
+# {{{ take/put
+
 @elwise_kernel_runner
 def _take(result, ary, indices):
     return elementwise.get_take_kernel(
@@ -636,10 +996,9 @@ def _take(result, ary, indices):
 
 
 def take(a, indices, out=None, queue=None):
+    queue = queue or a.queue
     if out is None:
-        out = Array(a.context, indices.shape, a.dtype, 
-                allocator=a.allocator,
-                queue=queue or a.queue)
+        out = Array(queue, indices.shape, a.dtype, allocator=a.allocator)
 
     assert len(indices.shape) == 1
     _take(out, a, indices, queue=queue)
@@ -663,7 +1022,7 @@ def multi_take(arrays, indices, out=None, queue=None):
     vec_count = len(arrays)
 
     if out is None:
-        out = [Array(context, queue, indices.shape, a_dtype, 
+        out = [Array(context, queue, indices.shape, a_dtype,
             allocator=a_allocator)
                 for i in range(vec_count)]
     else:
@@ -687,7 +1046,11 @@ def multi_take(arrays, indices, out=None, queue=None):
         if start_i + chunk_size > vec_count:
             knl = make_func_for_chunk_size(vec_count-start_i)
 
-        gs, ls = indices.get_sizes(queue)
+        gs, ls = indices.get_sizes(queue,
+                knl.get_work_group_info(
+                    cl.kernel_work_group_info.WORK_GROUP_SIZE,
+                    queue.device))
+
         knl(queue, gs, ls,
                 indices.data,
                 *([o.data for o in out[chunk_slice]]
@@ -713,8 +1076,7 @@ def multi_take_put(arrays, dest_indices, src_indices, dest_shape=None,
     vec_count = len(arrays)
 
     if out is None:
-        out = [Array(context, dest_shape, a_dtype, 
-            allocator=a_allocator, queue=queue)
+        out = [Array(queue, dest_shape, a_dtype, allocator=a_allocator)
                 for i in range(vec_count)]
     else:
         if a_dtype != single_valued(o.dtype for o in out):
@@ -756,7 +1118,11 @@ def multi_take_put(arrays, dest_indices, src_indices, dest_shape=None,
         if start_i + chunk_size > vec_count:
             knl = make_func_for_chunk_size(vec_count-start_i)
 
-        gs, ls = src_indices.get_sizes(queue)
+        gs, ls = src_indices.get_sizes(queue,
+                knl.get_work_group_info(
+                    cl.kernel_work_group_info.WORK_GROUP_SIZE,
+                    queue.device))
+
         knl(queue, gs, ls,
                 *([o.data for o in out[chunk_slice]]
                     + [dest_indices.data, src_indices.data]
@@ -809,7 +1175,11 @@ def multi_put(arrays, dest_indices, dest_shape=None, out=None, queue=None):
         if start_i + chunk_size > vec_count:
             knl = make_func_for_chunk_size(vec_count-start_i)
 
-        gs, ls = dest_indices.get_sizes(queue)
+        gs, ls = dest_indices.get_sizes(queue,
+                knl.get_work_group_info(
+                    cl.kernel_work_group_info.WORK_GROUP_SIZE,
+                    queue.device))
+
         knl(queue, gs, ls,
                 *([o.data for o in out[chunk_slice]]
                     + [dest_indices.data]
@@ -818,8 +1188,9 @@ def multi_put(arrays, dest_indices, dest_shape=None, out=None, queue=None):
 
     return out
 
+# }}}
 
-
+# {{{ conditionals
 
 @elwise_kernel_runner
 def _if_positive(result, criterion, then_, else_):
@@ -861,55 +1232,49 @@ def minimum(a, b, out=None, queue=None):
     return if_positive(a.mul_add(1, b, -1, queue=queue), b, a,
             queue=queue, out=out)
 
+# }}}
 
-
-
-# reductions ------------------------------------------------------------------
+# {{{ reductions
 _builtin_min = min
 _builtin_max = max
 
-if False:
-    # not yet
-    def sum(a, dtype=None, queue=None):
-        from pyopencl.reduction import get_sum_kernel
-        krnl = get_sum_kernel(dtype, a.dtype)
-        return krnl(a, queue=queue)
+def sum(a, dtype=None, queue=None):
+    from pyopencl.reduction import get_sum_kernel
+    krnl = get_sum_kernel(a.context, dtype, a.dtype)
+    return krnl(a, queue=queue)
 
-    def dot(a, b, dtype=None, queue=None):
-        from pyopencl.reduction import get_dot_kernel
-        krnl = get_dot_kernel(dtype, a.dtype, b.dtype)
-        return krnl(a, b, queue=queue)
+def dot(a, b, dtype=None, queue=None):
+    from pyopencl.reduction import get_dot_kernel
+    krnl = get_dot_kernel(a.context, dtype, a.dtype, b.dtype)
+    return krnl(a, b, queue=queue)
 
-    def subset_dot(subset, a, b, dtype=None, queue=None):
-        from pyopencl.reduction import get_subset_dot_kernel
-        krnl = get_subset_dot_kernel(dtype, a.dtype, b.dtype)
-        return krnl(subset, a, b, queue=queue)
+def subset_dot(subset, a, b, dtype=None, queue=None):
+    from pyopencl.reduction import get_subset_dot_kernel
+    krnl = get_subset_dot_kernel(a.context, dtype, subset.dtype, a.dtype, b.dtype)
+    return krnl(subset, a, b, queue=queue)
 
-    def _make_minmax_kernel(what):
-        def f(a, queue=None):
-            from pyopencl.reduction import get_minmax_kernel
-            krnl = get_minmax_kernel(what, a.dtype)
-            return krnl(a,  queue=queue)
+def _make_minmax_kernel(what):
+    def f(a, queue=None):
+        from pyopencl.reduction import get_minmax_kernel
+        krnl = get_minmax_kernel(a.context, what, a.dtype)
+        return krnl(a,  queue=queue)
 
-        return f
+    return f
 
-    min = _make_minmax_kernel("min")
-    max = _make_minmax_kernel("max")
-
-    def _make_subset_minmax_kernel(what):
-        def f(subset, a, queue=None):
-            from pyopencl.reduction import get_subset_minmax_kernel
-            import pyopencl.reduction
-            krnl = get_subset_minmax_kernel(what, a.dtype)
-            return krnl(subset, a,  queue=queue)
-
-        return f
-
-    subset_min = _make_subset_minmax_kernel("min")
-    subset_max = _make_subset_minmax_kernel("max")
+min = _make_minmax_kernel("min")
+max = _make_minmax_kernel("max")
 
+def _make_subset_minmax_kernel(what):
+    def f(subset, a, queue=None):
+        from pyopencl.reduction import get_subset_minmax_kernel
+        krnl = get_subset_minmax_kernel(a.context, what, a.dtype, subset.dtype)
+        return krnl(subset, a,  queue=queue)
 
+    return f
 
+subset_min = _make_subset_minmax_kernel("min")
+subset_max = _make_subset_minmax_kernel("max")
 
+# }}}
 
 # vim: foldmethod=marker
