@@ -32,25 +32,24 @@ class RanluxGenerator(object):
         self.max_work_items = max_work_items
 
         src = """
-            %s
+            %(defines)s
 
             #include <pyopencl-ranluxcl.cl>
 
             kernel void init_ranlux(unsigned seeds, global ranluxcl_state_t *ranluxcltab)
             {
-              ranluxcl_initialization(seeds, ranluxcltab);
+              if (get_global_id(0) < %(num_work_items)d)
+                ranluxcl_initialization(seeds, ranluxcltab);
             }
-            """ % self.generate_settings_defines()
+            """ % {
+                    "defines": self.generate_settings_defines(),
+                    "num_work_items": num_work_items
+                }
         prg = cl.Program(queue.context, src).build()
 
         # {{{ compute work group size
 
-        wg_size = prg.init_ranlux.get_work_group_info(
-                cl.kernel_work_group_info.WORK_GROUP_SIZE,
-                queue.device)
-
-        while num_work_items % wg_size != 0:
-            wg_size //= 2
+        wg_size = None
 
         import sys
         import platform
@@ -58,7 +57,7 @@ class RanluxGenerator(object):
                 and "Apple" in queue.device.platform.vendor
                 and platform.mac_ver()[0].startswith("10.7")
                 and queue.device.type == cl.device_type.CPU):
-            wg_size = 1
+            wg_size = (1,)
 
         self.wg_size = wg_size
 
@@ -67,7 +66,7 @@ class RanluxGenerator(object):
         self.state = cl_array.empty(queue, (num_work_items, 112), dtype=np.uint8)
         self.state.fill(17)
 
-        prg.init_ranlux(queue, (num_work_items,), (self.wg_size,), np.uint32(seed),
+        prg.init_ranlux(queue, (num_work_items,), self.wg_size, np.uint32(seed),
                 self.state.data)
 
     def generate_settings_defines(self, include_double_pragma=True):
@@ -93,6 +92,9 @@ class RanluxGenerator(object):
 
     @memoize_method
     def get_gen_kernel(self, dtype, flavor=""):
+        size_multiplier = 1
+        arg_dtype = dtype
+
         if dtype == np.float64:
             bits = 64
             c_type = "double"
@@ -101,6 +103,18 @@ class RanluxGenerator(object):
             bits = 32
             c_type = "float"
             rng_expr = "(shift + scale * gen)"
+        elif dtype == cl_array.vec.float2:
+            bits = 32
+            c_type = "float"
+            rng_expr = "(shift + scale * gen)"
+            size_multiplier = 2
+            arg_dtype = np.float32
+        elif dtype in [cl_array.vec.float3, cl_array.vec.float4]:
+            bits = 32
+            c_type = "float"
+            rng_expr = "(shift + scale * gen)"
+            size_multiplier = 4
+            arg_dtype = np.float32
         elif dtype == np.int32:
             assert flavor == ""
             bits = 32
@@ -132,6 +146,7 @@ class RanluxGenerator(object):
                 output_t scale,
                 output_t shift)
             {
+
               ranluxcl_state_t ranluxclstate;
               ranluxcl_download_seed(&ranluxclstate, ranluxcltab);
 
@@ -166,17 +181,18 @@ class RanluxGenerator(object):
 
         prg = cl.Program(self.context, src).build()
         knl = prg.generate
-        knl.set_scalar_arg_dtypes([None, None, np.uint64, dtype, dtype])
+        knl.set_scalar_arg_dtypes([None, None, np.uint64, arg_dtype, arg_dtype])
 
-        return knl
+        return knl, size_multiplier
 
     def fill_uniform(self, ary, a=0, b=1, queue=None):
         if queue is None:
             queue = ary.queue
 
-        self.get_gen_kernel(ary.dtype, "")(queue,
-                (self.num_work_items,), (self.wg_size,),
-                self.state.data, ary.data, ary.size,
+        knl, size_multiplier = self.get_gen_kernel(ary.dtype, "")
+        knl(queue,
+                (self.num_work_items,), None,
+                self.state.data, ary.data, ary.size*size_multiplier,
                 b-a, a)
 
     def uniform(self, *args, **kwargs):
@@ -192,9 +208,10 @@ class RanluxGenerator(object):
         if queue is None:
             queue = ary.queue
 
-        self.get_gen_kernel(ary.dtype, "norm")(queue,
-                (self.num_work_items,), (self.wg_size,),
-                self.state.data, ary.data, ary.size, sigma, mu)
+        knl, size_multiplier = self.get_gen_kernel(ary.dtype, "norm")
+        knl(queue,
+                (self.num_work_items,), self.wg_size,
+                self.state.data, ary.data, ary.size*size_multiplier, sigma, mu)
 
     def normal(self, *args, **kwargs):
         mu = kwargs.pop("mu", 0)
@@ -227,7 +244,7 @@ class RanluxGenerator(object):
         return prg.sync
 
     def synchronize(self, queue):
-        self.get_sync_kernel()(queue, (self.num_work_items,), None, self.state.data)
+        self.get_sync_kernel()(queue, (self.num_work_items,), self.wg_size, self.state.data)
 
 
 
