@@ -1,4 +1,5 @@
 #! /usr/bin/env python
+
 import numpy as np
 import numpy.linalg as la
 import sys
@@ -13,16 +14,17 @@ def have_cl():
         return False
 
 if have_cl():
-    import pyopencl.array as cl_array
     import pyopencl as cl
+    import pyopencl.array as cl_array
+    import pyopencl.tools as cl_tools
     from pyopencl.tools import pytest_generate_tests_for_pyopencl \
             as pytest_generate_tests
     from pyopencl.characterize import has_double_support
 
 
 @pytools.test.mark_test.opencl
-def test_pow_array(ctx_getter):
-    context = ctx_getter()
+def test_pow_array(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     a = np.array([1, 2, 3, 4, 5]).astype(np.float32)
@@ -36,8 +38,8 @@ def test_pow_array(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_pow_number(ctx_getter):
-    context = ctx_getter()
+def test_pow_number(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
@@ -48,28 +50,137 @@ def test_pow_number(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_abs(ctx_getter):
-    context = ctx_getter()
+def test_absrealimag(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
-    a = -cl_array.arange(queue, 111, dtype=np.float32)
-    res = a.get()
+    def real(x): return x.real
+    def imag(x): return x.imag
+    def conj(x): return x.conj()
 
-    for i in range(111):
-        assert res[i] <= 0
+    n = 111
+    for func in [abs, real, imag, conj]:
+        for dtype in [np.int32, np.float32, np.complex64]:
+            print(func, dtype)
+            a = -make_random_array(queue, dtype, n)
 
-    a = abs(a)
+            host_res = func(a.get())
+            dev_res = func(a).get()
 
-    res = a.get()
+            correct = np.allclose(dev_res, host_res)
+            if not correct:
+                print(dev_res)
+                print(host_res)
+                print(dev_res-host_res)
+            assert correct
 
-    for i in range(111):
-        assert abs(res[i]) >= 0
-        assert res[i] == i
 
+TO_REAL = {
+        np.dtype(np.complex64): np.float32,
+        np.dtype(np.complex128): np.float64
+        }
+
+def make_random_array(queue, dtype, size):
+    from pyopencl.clrandom import rand
+
+    dtype = np.dtype(dtype)
+    if dtype.kind == "c":
+        real_dtype = TO_REAL[dtype]
+        return (rand(queue, shape=(size,), dtype=real_dtype).astype(dtype)
+                + dtype.type(1j)
+                * rand(queue, shape=(size,), dtype=real_dtype).astype(dtype))
+    else:
+        return rand(queue, shape=(size,), dtype=dtype)
 
 @pytools.test.mark_test.opencl
-def test_len(ctx_getter):
-    context = ctx_getter()
+def test_basic_complex(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    from pyopencl.clrandom import rand
+
+    size = 500
+
+    ary =  (rand(queue, shape=(size,), dtype=np.float32).astype(np.complex64)
+            + 1j* rand(queue, shape=(size,), dtype=np.float32).astype(np.complex64))
+    c = np.complex64(5+7j)
+
+    host_ary = ary.get()
+    assert la.norm((c*ary).get() - c*host_ary) < 1e-5 * la.norm(host_ary)
+
+@pytools.test.mark_test.opencl
+def test_mix_complex(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    size = 10
+
+    dtypes = [
+            (np.float32, np.complex64),
+            #(np.int32, np.complex64),
+            ]
+
+    if has_double_support(context.devices[0]):
+        dtypes.extend([
+            (np.float32, np.float64),
+            (np.float32, np.complex128),
+            (np.float64, np.complex64),
+            (np.float64, np.complex128),
+            ])
+
+    from operator import add, mul, sub, truediv
+    for op in [add, sub, mul, truediv, pow]:
+        for dtype_a0, dtype_b0 in dtypes:
+            for dtype_a, dtype_b in [
+                    (dtype_a0, dtype_b0),
+                    (dtype_b0, dtype_a0),
+                    ]:
+                for is_scalar_a, is_scalar_b in [
+                        (False, False),
+                        (False, True),
+                        (True, False),
+                        ]:
+                    if is_scalar_a:
+                        ary_a = make_random_array(queue, dtype_a, 1).get()[0]
+                        host_ary_a = ary_a
+                    else:
+                        ary_a = make_random_array(queue, dtype_a, size)
+                        host_ary_a = ary_a.get()
+
+                    if is_scalar_b:
+                        ary_b = make_random_array(queue, dtype_b, 1).get()[0]
+                        host_ary_b = ary_b
+                    else:
+                        ary_b = make_random_array(queue, dtype_b, size)
+                        host_ary_b = ary_b.get()
+
+                    print(op, dtype_a, dtype_b, is_scalar_a, is_scalar_b)
+                    dev_result = op(ary_a, ary_b).get()
+                    host_result = op(host_ary_a, host_ary_b)
+
+                    if host_result.dtype != dev_result.dtype:
+                        # This appears to be a numpy bug, where we get
+                        # served a Python complex that is really a
+                        # smaller numpy complex.
+
+                        print("HOST_DTYPE: %s DEV_DTYPE: %s" % (
+                                host_result.dtype, dev_result.dtype))
+
+                        dev_result = dev_result.astype(host_result.dtype)
+
+                    err = la.norm(host_result-dev_result)/la.norm(host_result)
+                    print(err)
+                    correct = err < 1e-5
+                    if not correct:
+                        print(host_result)
+                        print(dev_result)
+                        print(host_result - dev_result)
+
+                    assert correct
+
+@pytools.test.mark_test.opencl
+def test_len(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
@@ -78,29 +189,30 @@ def test_len(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_multiply(ctx_getter):
+def test_multiply(ctx_factory):
     """Test the muliplication of an array with a scalar. """
 
-    context = ctx_getter()
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     for sz in [10, 50000]:
         for dtype, scalars in [
             (np.float32, [2]),
+            (np.complex64, [2j]),
             ]:
             for scalar in scalars:
-                a = np.arange(sz).astype(dtype)
-                a_gpu = cl_array.to_device(queue, a)
-                a_doubled = (scalar * a_gpu).get()
+                a_gpu = make_random_array(queue, dtype, sz)
+                a = a_gpu.get()
+                a_mult = (scalar * a_gpu).get()
 
-                assert (a * scalar == a_doubled).all()
+                assert (a * scalar == a_mult).all()
 
 
 @pytools.test.mark_test.opencl
-def test_multiply_array(ctx_getter):
+def test_multiply_array(ctx_factory):
     """Test the multiplication of two arrays."""
 
-    context = ctx_getter()
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
@@ -114,10 +226,10 @@ def test_multiply_array(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_addition_array(ctx_getter):
+def test_addition_array(ctx_factory):
     """Test the addition of two arrays."""
 
-    context = ctx_getter()
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
@@ -128,10 +240,10 @@ def test_addition_array(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_addition_scalar(ctx_getter):
+def test_addition_scalar(ctx_factory):
     """Test the addition of an array and a scalar."""
 
-    context = ctx_getter()
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
@@ -142,14 +254,14 @@ def test_addition_scalar(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_substract_array(ctx_getter):
+def test_substract_array(ctx_factory):
     """Test the substraction of two arrays."""
     #test data
     a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
     b = np.array([10, 20, 30, 40, 50,
                   60, 70, 80, 90, 100]).astype(np.float32)
 
-    context = ctx_getter()
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     a_gpu = cl_array.to_device(queue, a)
@@ -163,10 +275,10 @@ def test_substract_array(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_substract_scalar(ctx_getter):
+def test_substract_scalar(ctx_factory):
     """Test the substraction of an array and a scalar."""
 
-    context = ctx_getter()
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     #test data
@@ -183,10 +295,10 @@ def test_substract_scalar(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_divide_scalar(ctx_getter):
+def test_divide_scalar(ctx_factory):
     """Test the division of an array and a scalar."""
 
-    context = ctx_getter()
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     a = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(np.float32)
@@ -200,10 +312,10 @@ def test_divide_scalar(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_divide_array(ctx_getter):
+def test_divide_array(ctx_factory):
     """Test the division of an array and a scalar. """
 
-    context = ctx_getter()
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     #test data
@@ -221,27 +333,51 @@ def test_divide_array(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_random(ctx_getter):
-    context = ctx_getter()
+def test_random(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
-    from pyopencl.clrandom import rand as clrand
+    from pyopencl.clrandom import RanluxGenerator
 
     if has_double_support(context.devices[0]):
         dtypes = [np.float32, np.float64]
     else:
         dtypes = [np.float32]
 
+    gen = RanluxGenerator(queue, 5120)
+
+    for ary_size in [300, 301, 302, 303, 10007]:
+        for dtype in dtypes:
+            ran = cl_array.zeros(queue, ary_size, dtype)
+            gen.fill_uniform(ran)
+            assert (0 < ran.get()).all()
+            assert (ran.get() < 1).all()
+
+            gen.synchronize(queue)
+
+            ran = cl_array.zeros(queue, ary_size, dtype)
+            gen.fill_uniform(ran, a=4, b=7)
+            assert (4 < ran.get()).all()
+            assert (ran.get() < 7).all()
+
+            ran = gen.normal(queue, (10007,), dtype, mu=4, sigma=3)
+
+    dtypes = [np.int32]
     for dtype in dtypes:
-        a = clrand(context, queue, (10, 100), dtype=dtype).get()
+        ran = gen.uniform(queue, (10000007,), dtype, a=200, b=300)
+        assert (200 <= ran.get()).all()
+        assert (ran.get() < 300).all()
+        #from matplotlib import pyplot as pt
+        #pt.hist(ran.get())
+        #pt.show()
+
+
 
-        assert (0 <= a).all()
-        assert (a < 1).all()
 
 
 @pytools.test.mark_test.opencl
-def test_nan_arithmetic(ctx_getter):
-    context = ctx_getter()
+def test_nan_arithmetic(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     def make_nan_contaminated_vector(size):
@@ -266,14 +402,14 @@ def test_nan_arithmetic(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_elwise_kernel(ctx_getter):
-    context = ctx_getter()
+def test_elwise_kernel(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     from pyopencl.clrandom import rand as clrand
 
-    a_gpu = clrand(context, queue, (50,), np.float32)
-    b_gpu = clrand(context, queue, (50,), np.float32)
+    a_gpu = clrand(queue, (50,), np.float32)
+    b_gpu = clrand(queue, (50,), np.float32)
 
     from pyopencl.elementwise import ElementwiseKernel
     lin_comb = ElementwiseKernel(context,
@@ -288,16 +424,16 @@ def test_elwise_kernel(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_elwise_kernel_with_options(ctx_getter):
+def test_elwise_kernel_with_options(ctx_factory):
     from pyopencl.clrandom import rand as clrand
     from pyopencl.elementwise import ElementwiseKernel
 
-    context = ctx_getter()
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
-    in_gpu = clrand(context, queue, (50,), np.float32)
+    in_gpu = clrand(queue, (50,), np.float32)
 
-    options = ['-DADD_ONE']
+    options = ['-D', 'ADD_ONE']
     add_one = ElementwiseKernel(
         context,
         "float* out, const float *in",
@@ -320,8 +456,8 @@ def test_elwise_kernel_with_options(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_take(ctx_getter):
-    context = ctx_getter()
+def test_take(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     idx = cl_array.arange(queue, 0, 200000, 2, dtype=np.uint32)
@@ -331,8 +467,8 @@ def test_take(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_arange(ctx_getter):
-    context = ctx_getter()
+def test_arange(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     n = 5000
@@ -341,8 +477,8 @@ def test_arange(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_reverse(ctx_getter):
-    context = ctx_getter()
+def test_reverse(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     n = 5000
@@ -355,13 +491,13 @@ def test_reverse(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_sum(ctx_getter):
-    context = ctx_getter()
+def test_sum(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     from pyopencl.clrandom import rand as clrand
 
-    a_gpu = clrand(context, queue, (200000,), np.float32)
+    a_gpu = clrand(queue, (200000,), np.float32)
     a = a_gpu.get()
 
     sum_a = np.sum(a)
@@ -371,8 +507,8 @@ def test_sum(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_minmax(ctx_getter):
-    context = ctx_getter()
+def test_minmax(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     from pyopencl.clrandom import rand as clrand
@@ -384,7 +520,7 @@ def test_minmax(ctx_getter):
 
     for what in ["min", "max"]:
         for dtype in dtypes:
-            a_gpu = clrand(context, queue, (200000,), dtype)
+            a_gpu = clrand(queue, (200000,), dtype)
             a = a_gpu.get()
 
             op_a = getattr(np, what)(a)
@@ -394,8 +530,8 @@ def test_minmax(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_subset_minmax(ctx_getter):
-    context = ctx_getter()
+def test_subset_minmax(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     from pyopencl.clrandom import rand as clrand
@@ -410,7 +546,7 @@ def test_subset_minmax(ctx_getter):
         dtypes = [np.float32, np.int32]
 
     for dtype in dtypes:
-        a_gpu = clrand(context, queue, (l_a,), dtype)
+        a_gpu = clrand(queue, (l_a,), dtype)
         a = a_gpu.get()
 
         meaningful_indices_gpu = cl_array.zeros(
@@ -434,14 +570,14 @@ def test_subset_minmax(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_dot(ctx_getter):
-    context = ctx_getter()
+def test_dot(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     from pyopencl.clrandom import rand as clrand
-    a_gpu = clrand(context, queue, (200000,), np.float32)
+    a_gpu = clrand(queue, (200000,), np.float32)
     a = a_gpu.get()
-    b_gpu = clrand(context, queue, (200000,), np.float32)
+    b_gpu = clrand(queue, (200000,), np.float32)
     b = b_gpu.get()
 
     dot_ab = np.dot(a, b)
@@ -453,11 +589,11 @@ def test_dot(ctx_getter):
 
 if False:
     @pytools.test.mark_test.opencl
-    def test_slice(ctx_getter):
+    def test_slice(ctx_factory):
         from pyopencl.clrandom import rand as clrand
 
         l = 20000
-        a_gpu = clrand(context, queue, (l,))
+        a_gpu = clrand(queue, (l,))
         a = a_gpu.get()
 
         from random import randrange
@@ -472,15 +608,15 @@ if False:
 
 
 @pytools.test.mark_test.opencl
-def test_if_positive(ctx_getter):
-    context = ctx_getter()
+def test_if_positive(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     from pyopencl.clrandom import rand as clrand
 
     l = 20000
-    a_gpu = clrand(context, queue, (l,), np.float32)
-    b_gpu = clrand(context, queue, (l,), np.float32)
+    a_gpu = clrand(queue, (l,), np.float32)
+    b_gpu = clrand(queue, (l,), np.float32)
     a = a_gpu.get()
     b = b_gpu.get()
 
@@ -495,8 +631,8 @@ def test_if_positive(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_take_put(ctx_getter):
-    context = ctx_getter()
+def test_take_put(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     for n in [5, 17, 333]:
@@ -517,8 +653,8 @@ def test_take_put(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_astype(ctx_getter):
-    context = ctx_getter()
+def test_astype(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     from pyopencl.clrandom import rand as clrand
@@ -526,7 +662,7 @@ def test_astype(ctx_getter):
     if not has_double_support(context.devices[0]):
         return
 
-    a_gpu = clrand(context, queue, (2000,), dtype=np.float32)
+    a_gpu = clrand(queue, (2000,), dtype=np.float32)
 
     a = a_gpu.get().astype(np.float64)
     a2 = a_gpu.astype(np.float64).get()
@@ -534,7 +670,7 @@ def test_astype(ctx_getter):
     assert a2.dtype == np.float64
     assert la.norm(a - a2) == 0, (a, a2)
 
-    a_gpu = clrand(context, queue, (2000,), dtype=np.float64)
+    a_gpu = clrand(queue, (2000,), dtype=np.float64)
 
     a = a_gpu.get().astype(np.float32)
     a2 = a_gpu.astype(np.float32).get()
@@ -543,15 +679,39 @@ def test_astype(ctx_getter):
     assert la.norm(a - a2) / la.norm(a) < 1e-7
 
 
+def summarize_error(obtained, desired, orig, thresh=1e-5):
+    err = obtained - desired
+    ok_count = 0
+
+    entries = []
+    for i, val in enumerate(err):
+        if abs(val) > thresh:
+            if ok_count:
+                entries.append("<%d ok>" % ok_count)
+                ok_count = 0
+
+            entries.append("%r (want: %r, diff: %r, orig: %r)" % (obtained[i], desired[i],
+                obtained[i]-desired[i], orig[i]))
+        else:
+            ok_count += 1
+
+    if ok_count:
+        entries.append("<%d ok>" % ok_count)
+
+    return " ".join(entries)
+
 @pytools.test.mark_test.opencl
-def test_scan(ctx_getter):
-    context = ctx_getter()
+def test_scan(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     from pyopencl.scan import InclusiveScanKernel, ExclusiveScanKernel
 
     dtype = np.int32
-    for cls in [InclusiveScanKernel, ExclusiveScanKernel]:
+    for cls in [
+            InclusiveScanKernel,
+            ExclusiveScanKernel
+            ]:
         knl = cls(context, dtype, "a+b", "0")
 
         for n in [
@@ -559,25 +719,33 @@ def test_scan(ctx_getter):
             2 ** 20 - 2 ** 18,
             2 ** 20 - 2 ** 18 + 5,
             2 ** 10 + 5,
-            2 ** 20 + 5,
-            2 ** 20, 2 ** 24]:
+            2 ** 20 + 1,
+            2 ** 20, 2 ** 24
+            ]:
+
             host_data = np.random.randint(0, 10, n).astype(dtype)
             dev_data = cl_array.to_device(queue, host_data)
 
+            assert (host_data == dev_data.get()).all() # /!\ fails on Nv GT2?? for some drivers
+
             knl(dev_data)
 
             desired_result = np.cumsum(host_data, axis=0)
             if cls is ExclusiveScanKernel:
                 desired_result -= host_data
 
-            assert (dev_data.get() == desired_result).all()
+            is_ok = (dev_data.get() == desired_result).all()
+            if 0 and not is_ok:
+                print(summarize_error(dev_data.get(), desired_result, host_data))
+
+            assert is_ok
             from gc import collect
             collect()
 
 
 @pytools.test.mark_test.opencl
-def test_stride_preservation(ctx_getter):
-    context = ctx_getter()
+def test_stride_preservation(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     A = np.random.rand(3, 3)
@@ -589,8 +757,8 @@ def test_stride_preservation(ctx_getter):
 
 
 @pytools.test.mark_test.opencl
-def test_vector_fill(ctx_getter):
-    context = ctx_getter()
+def test_vector_fill(ctx_factory):
+    context = ctx_factory()
     queue = cl.CommandQueue(context)
 
     a_gpu = cl_array.Array(queue, 100, dtype=cl_array.vec.float4)
@@ -600,6 +768,115 @@ def test_vector_fill(ctx_getter):
 
     a_gpu = cl_array.zeros(queue, 100, dtype=cl_array.vec.float4)
 
+@pytools.test.mark_test.opencl
+def test_mem_pool_with_arrays(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+    mem_pool = cl_tools.MemoryPool(cl_tools.CLAllocator(context))
+
+    a_dev = cl_array.arange(queue, 2000, dtype=np.float32, allocator=mem_pool)
+    b_dev = cl_array.to_device(queue, np.arange(2000), allocator=mem_pool) + 4000
+
+    result = cl_array.dot(a_dev, b_dev)
+    assert a_dev.allocator is mem_pool
+    assert b_dev.allocator is mem_pool
+    assert result.allocator is mem_pool
+
+@pytools.test.mark_test.opencl
+def test_view(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    a = np.arange(128).reshape(8, 16).astype(np.float32)
+    a_dev = cl_array.to_device(queue, a)
+
+    # same dtype
+    view = a_dev.view()
+    assert view.shape == a_dev.shape and view.dtype == a_dev.dtype
+
+    # larger dtype
+    view = a_dev.view(np.complex64)
+    assert view.shape == (8, 8) and view.dtype == np.complex64
+
+    # smaller dtype
+    view = a_dev.view(np.int16)
+    assert view.shape == (8, 32) and view.dtype == np.int16
+
+mmc_dtype = np.dtype([
+    ("cur_min", np.int32),
+    ("cur_max", np.int32),
+    ("pad", np.int32),
+    ])
+
+from pyopencl.tools import register_dtype
+register_dtype(mmc_dtype, "minmax_collector", alias_ok=True)
+register_dtype(mmc_dtype, "minmax_collector", alias_ok=True)
+
+@pytools.test.mark_test.opencl
+def test_struct_reduce(ctx_factory):
+    context = ctx_factory()
+    queue = cl.CommandQueue(context)
+
+    preamble = r"""//CL//
+    struct minmax_collector
+    {
+        int cur_min;
+        int cur_max;
+        // Workaround for OS X Lion GPU CL. Mystifying.
+        int pad;
+    };
+
+    typedef struct minmax_collector minmax_collector;
+
+    minmax_collector mmc_neutral()
+    {
+        // FIXME: needs infinity literal in real use, ok here
+        minmax_collector result;
+        result.cur_min = 1<<30;
+        result.cur_max = -(1<<30);
+        return result;
+    }
+
+    minmax_collector mmc_from_scalar(float x)
+    {
+        minmax_collector result;
+        result.cur_min = x;
+        result.cur_max = x;
+        return result;
+    }
+
+    minmax_collector agg_mmc(minmax_collector a, minmax_collector b)
+    {
+        minmax_collector result = a;
+        if (b.cur_min < result.cur_min)
+            result.cur_min = b.cur_min;
+        if (b.cur_max > result.cur_max)
+            result.cur_max = b.cur_max;
+        return result;
+    }
+
+    """
+
+
+    from pyopencl.clrandom import rand as clrand
+    a_gpu = clrand(queue, (20000,), dtype=np.int32, a=0, b=10**6)
+    a = a_gpu.get()
+
+    from pyopencl.reduction import ReductionKernel
+    red = ReductionKernel(context, mmc_dtype,
+            neutral="mmc_neutral()",
+            reduce_expr="agg_mmc(a, b)", map_expr="mmc_from_scalar(x[i])",
+            arguments="__global int *x", preamble=preamble)
+
+    minmax = red(a_gpu).get()
+    #print minmax["cur_min"], minmax["cur_max"]
+    #print np.min(a), np.max(a)
+
+    assert abs(minmax["cur_min"] - np.min(a)) < 1e-5
+    assert abs(minmax["cur_max"] - np.max(a)) < 1e-5
+
+
+
 
 if __name__ == "__main__":
     # make sure that import failures get reported, instead of skipping the
@@ -612,3 +889,5 @@ if __name__ == "__main__":
     else:
         from py.test.cmdline import main
         main([__file__])
+
+# vim: filetype=pyopencl
